from .drivers import current_driver
from .data_variable import DataVariable
from .support_data_variable import SupportDataVariable
import re
from typing import List

DEPEND_REGEX = re.compile("DEPEND_\\d")


def _get_attributes(cdf: object, var: str):
    attrs = {}
    for attr in cdf.variable_attributes(var):
        if attr.endswith("_PTR") or attr[:-1].endswith("_PTR_"):
            value = cdf.values(cdf.variable_attribute_value(var, attr))
            if hasattr(value, 'tolist'):
                attrs[attr] = value.tolist()
            else:
                attrs[attr] = value
        else:
            attrs[attr] = cdf.variable_attribute_value(var, attr)
    return attrs


def _get_axis(cdf: object, var: str):
    if cdf.is_char(var):
        return None
    return SupportDataVariable(name=var, values=cdf.values(var), attributes=_get_attributes(cdf, var))


def _get_axes(cdf: object, var: str):
    attrs = sorted(filter(lambda attr: DEPEND_REGEX.match(attr), cdf.variable_attributes(var)))
    unix_time_name = cdf.variable_attribute_value(var, "DEPEND_TIME")
    axes = list(map(lambda attr: _get_axis(cdf, cdf.variable_attribute_value(var, attr)), attrs))
    if unix_time_name is not None:
        unix_time = _get_axis(cdf, unix_time_name)
        if len(unix_time) > len(axes[0]):
            unix_time.values = (unix_time.values * 1e9).astype('<M8[ns]')
            axes[0] = unix_time
            Warning(
                "Non compliant CDF file, swapping DEPEND_0 with DEPEND_TIME, if you think this is a bug report it here: https://github.com/SciQLop/PyISTP/issues")
    return axes


def _get_labels(attributes) -> List[str]:
    if 'LABL_PTR_1' in attributes:
        return attributes['LABL_PTR_1']
    if 'LABLAXIS' in attributes:
        return [attributes['LABLAXIS']]


def _load_data_var(cdf: object, var: str) -> DataVariable or None:
    axes = _get_axes(cdf, var)
    attributes = _get_attributes(cdf, var)
    labels = _get_labels(attributes)
    if None in axes:
        return None
    return DataVariable(name=var, values=cdf.values(var), attributes=attributes, axes=axes, labels=labels)


class ISTPLoaderImpl:
    cdf = None

    def __init__(self, file=None, buffer=None):
        self.cdf = current_driver(file or buffer)
        self.data_variables = []
        self._update_data_vars_lis()

    def attributes(self):
        return self.cdf.attributes()

    def attribute(self, key):
        return self.cdf.attribute(key)

    def _update_data_vars_lis(self):
        if self.cdf:
            self.data_variables = []
            for var in self.cdf.variables():
                var_attrs = self.cdf.variable_attributes(var)
                var_type = self.cdf.variable_attribute_value(var, 'VAR_TYPE')
                if var_type == 'data' and not self.cdf.is_char(var):
                    self.data_variables.append(var)

    def data_variable(self, var_name):
        return _load_data_var(self.cdf, var_name)



import pycdfpp


class Driver:
    def __init__(self, file):
        self.cdf = pycdfpp.load(file)

    def attributes(self):
        if self.cdf:
            return self.cdf.attributes.keys()
        return []

    def attribute(self, key):
        if self.cdf:
            return self.cdf.attributes[key]
        return None

    def variables(self):
        if self.cdf:
            return list(self.cdf)
        return []

    def is_char(self,var):
        return self.cdf[var].type == pycdfpp.CDF_CHAR

    def variable_attributes(self, var):
        if self.cdf:
            return self.cdf[var].attributes.keys()
        return []

    def variable_attribute_value(self, var, attr):
        if self.cdf and var in self.cdf and attr in self.cdf[var].attributes:
            return self.cdf[var].attributes[attr][0]
        return None

    def values(self, var):
        v = self.cdf[var]
        if v.type in (pycdfpp.CDF_EPOCH, pycdfpp.CDF_EPOCH16, pycdfpp.CDF_TIME_TT2000):
            return pycdfpp.to_datetime64(v)
        return v.values



from typing import List


class ISTPLoader:

    def __init__(self, file=None, buffer=None):
        from ._impl import ISTPLoaderImpl
        self._impl = ISTPLoaderImpl(file=file, buffer=buffer)

    def attributes(self):
        return self._impl.attributes()

    def attribute(self, key):
        return self._impl.attribute(key)

    def data_variables(self) -> List[str]:
        return self._impl.data_variables

    def data_variable(self, var_name) -> List[str]:
        return self._impl.data_variable(var_name)



from spacepy import pycdf
import numpy as np
from tempfile import NamedTemporaryFile


class Driver:
    def __init__(self, file):
        if type(file) is bytes:
            self._tmpfile = NamedTemporaryFile()
            self._tmpfile.write(file)
            self._tmpfile.flush()
            self.cdf = pycdf.CDF(self._tmpfile.name)
        else:
            self.cdf = pycdf.CDF(file)

    def attributes(self):
        if self.cdf:
            return self.cdf.attrs.keys()
        return []

    def attribute(self, key):
        if self.cdf:
            return self.cdf.attrs[key]
        return None

    def variables(self):
        if self.cdf:
            return self.cdf.keys()
        return []

    def variable_attributes(self, var):
        if self.cdf:
            return self.cdf[var].attrs.keys()
        return []

    def is_char(self,var):
        return self.cdf[var].type() == pycdf.const.CDF_CHAR

    def variable_attribute_value(self, var, attr):
        if self.cdf and var in self.cdf and attr in self.cdf[var].attrs:
            return self.cdf[var].attrs[attr]
        return None

    def values(self, var):
        v = self.cdf[var]
        if v.type() in (pycdf.const.CDF_EPOCH, pycdf.const.CDF_EPOCH16, pycdf.const.CDF_TIME_TT2000):
            return np.vectorize(np.datetime64)(v[:])
        return v[:]



class DataVariable:
    __slots__ = ("name", "values", 'attributes', 'axes', 'labels')

    def __init__(self, name, values, attributes, axes, labels=None):
        self.name = name
        self.values = values
        self.attributes = attributes
        self.axes = axes or []
        self.labels = labels

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return f"""DataVariable: {self.name}
Attributes:
{self.attributes}
Axes:
{self.axes}\n"""



class SupportDataVariable:
    __slots__ = ("name", "values", 'attributes')

    def __init__(self, name, values, attributes):
        self.name = name
        self.values = values
        self.attributes = attributes

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return f"""SupportDataVariable: {self.name}
Attributes:
{self.attributes}\n"""



import logging, os
import importlib

log = logging.getLogger(__name__)


def _load_cdf_lib():
    available_libs = ["pycdfpp", "spacepy"]
    try_first_lib = os.environ.get("PYISTP_CDFLIB", "pycdfpp")
    available_libs.remove(try_first_lib)
    available_libs.insert(0, try_first_lib)

    for driver in available_libs:
        try:
            mod = importlib.import_module(f"pyistp.drivers.{driver}", "*")
            log.info(f"Using {driver}")
            return mod.Driver
        except ImportError:
            log.info(f"Failed to load {driver}, trying next available driver")
    raise ImportError("Can't import any compatible CDF library")


current_driver = _load_cdf_lib()
